/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package ij3d.image3d;

import ij.IJ;
import ij3d.geom.Voxel3DSeg;
import ij3d.utils.ArrayUtil;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;

/**
Copyright (C) Thomas Boudier

License:
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 *
 * @author thomas
 */
public class Segment3DSpots {

    IntImage3D image;
    IntImage3D seeds; // positions of seeds
    IntImage3D watershedImage = null;
    int globalThreshold = -1;
    int localThreshold = -1;
    ArrayList voxels = null;
    final int NO_LABEL = 0;
    final int BORDER = 1;
    final int FIRST_LABEL = 2;

    public IntImage3D getWatershedImage() {
        if (watershedImage == null) {
            this.watershed();
        }
        return watershedImage;
    }

    public int getLocalThreshold() {
        return localThreshold;
    }

    public void setLocalThreshold(int localThreshold) {
        this.localThreshold = localThreshold;
    }

    public int getGlobalThreshold() {
        return globalThreshold;
    }

    public void setGlobalThreshold(int globalThreshold) {
        this.globalThreshold = globalThreshold;
    }

    public Segment3DSpots(IntImage3D image, IntImage3D seeds) {
        this.image = image;
        this.seeds = seeds;
    }

    public IntImage3D getImage() {
        return image;
    }

    public void setImage(IntImage3D image) {
        this.image = image;
    }

    public IntImage3D getSeeds() {
        return seeds;
    }

    public void setSeeds(IntImage3D seeds) {
        this.seeds = seeds;
    }

    private void watershed() {
        this.createArrayList();
        //IJ.log("Nb voxels = " + voxels.size());
        IJ.showStatus("Sorting ...");
        Collections.shuffle(voxels);
        Collections.sort(voxels);

        Iterator it;
        Voxel3DSeg voxel;
        ArrayUtil neigh;
        int max, max2;
        int px, py, pz;

        ArrayList voxels2;

        boolean loop = true;
        int sx = image.getSizex();
        int sy = image.getSizey();
        int sz = image.getSizez();
        int MaxIterations = (int) Math.sqrt(sx * sx + sy * sy + sz * sz);
        //IJ.log("ite=" + MaxIterations);
        int ite = 0;
        int nb = voxels.size();

        while ((loop) && (ite < MaxIterations)) {
            IJ.showStatus("Watershed " + ite + " (" + voxels.size() + ")");
            loop = false;
            ite++;
            voxels2 = new ArrayList();
            for (it = voxels.iterator(); it.hasNext();) {
                voxel = (Voxel3DSeg) it.next();
                px = (int) voxel.getX();
                py = (int) voxel.getY();
                pz = (int) voxel.getZ();
                if (watershedImage.getPixel(px, py, pz) == NO_LABEL) {
                    neigh = watershedImage.getNeighborhoodCross3D(px, py, pz);
                    max = (int) neigh.getMaximum();
                    if ((max == NO_LABEL) || (max == BORDER)) {
                        loop = true;
                        voxels2.add(voxel);
                    } else {
                        max2 = (int) neigh.getMaximumBelow(max);
                        if ((max2 == NO_LABEL) || (max == BORDER)) {
                            watershedImage.putPixel(px, py, pz, max);
                        } else {
                            watershedImage.putPixel(px, py, pz, BORDER);
                        }
                    }
                }
            }
            if (nb == voxels2.size()) {
                loop = false;
            } else {
                voxels = voxels2;
                nb = voxels.size();
            }
        }
    }

    private void createArrayList() {
        voxels = new ArrayList();
        final int sx = image.getSizex();
        final int sy = image.getSizey();
        final int sz = image.getSizez();
        final int th = this.globalThreshold;


        // watershedImage
        this.watershedImage = new IntImage3D(sx, sy, sz);

        int pix, se;
        int c = FIRST_LABEL;

        for (int z = 0; z < sz; z++) {
            IJ.showStatus("Processsing " + z);
            for (int y = 0; y < sy; y++) {
                for (int x = 0; x < sx; x++) {
                    pix = image.getPixel(x, y, z);
                    se = seeds.getPixel(x, y, z);
                    if (pix > th) {
                        if (se > 0) {
                            watershedImage.putPixel(x, y, z, c);
                            voxels.add(new Voxel3DSeg(x, y, z, pix, c));
                            c++;
                        } else {
                            voxels.add(new Voxel3DSeg(x, y, z, pix, NO_LABEL));
                            watershedImage.putPixel(x, y, z, NO_LABEL);
                        }
                    } else {
                        watershedImage.putPixel(x, y, z, NO_LABEL);
                    }
                }
            }
        }
    }

    /**
     *  Local mean for segmentation
     *
     * @param  xc      x coordinate of pixel
     * @param  yc      y coordinate of pixel
     * @param  zc      z coordinate of pixel
     * @param  radius  radius to compute localMean mean
     * @param  st      the stack
     * @return         the local mean
     */
    public double localMean(int xc, int yc, int zc, float rad0, float rad1, float rad2, double weight) {
        double mspot = 0, mback = 0;
        int sx = image.getSizex();
        int sy = image.getSizey();
        int sz = image.getSizez();

        if (watershedImage == null) {
            // no watershed image
            // central value
            ArrayUtil neigh = image.getNeighborhood(xc, yc, zc, rad0, rad0, rad0);
            mspot = neigh.getMean();
            // background value
            neigh = image.getNeighborhoodLayer(xc, yc, zc, rad1, rad2);
            mback = neigh.getMean();
        } else {
            // take into account watershed image
            double rad02 = rad0 * rad0;
            double rad12 = rad1 * rad1;
            double rad22 = rad2 * rad2;
            int zd, zf, yd, yf, xd, xf;
            int nbspot = 0;
            int nbback = 0;
            int sumspot = 0;
            int sumback = 0;
            double dist = 0;
            int pix, water;
            int waterc = watershedImage.getPixel(xc, yc, zc);

            // compue bounding
            zd = (int) (zc - rad2);
            if (zd < 0) {
                zd = 0;
            }
            zf = (int) (zc + rad2);
            if (zf >= sz) {
                zf = sz;
            }
            yd = (int) (yc - rad2);
            if (yd < 0) {
                yd = 0;
            }
            yf = (int) (yc + rad2);
            if (yf >= sy) {
                yf = sy;
            }
            xd = (int) (xc - rad2);
            if (xd < 0) {
                xd = 0;
            }
            xf = (int) (xc + rad2);
            if (xf >= sx) {
                xf = sx;
            }

            for (int z = zd; z < zf; z++) {
                for (int y = yd; y < yf; y++) {
                    for (int x = xd; x < xf; x++) {
                        dist = (x - xc) * (x - xc) + (y - yc) * (y - yc) + (z - zc) * (z - zc);
                        if (dist <= rad02) {
                            water = watershedImage.getPixel(x, y, z);
                            if (water == waterc) {
                                pix = image.getPixel(x, y, z);
                                sumspot += pix;
                                nbspot++;
                            }
                        } else if ((dist >= rad12) && (dist <= rad22)) {
                            water = watershedImage.getPixel(x, y, z);
                            if (water == waterc) {
                                pix = image.getPixel(x, y, z);
                                sumback += pix;
                                nbback++;
                            }
                        }
                    }
                }
            }
            // test nb minimum of pixels in spot and background
            if (nbspot > 1) {
                mspot = sumspot / nbspot;
            } else {
                return -1;
            }
            if (nbback > 1) {
                mback = sumback / nbback;
            } else {
                return -1;
            }
        }

        return (mspot * weight + (1.0 - weight) * mback);
    }
}
